#include <cstdio>
#include <vector>
#include <list>
#include <pair.h>

using namespace std;

#define HISTORYDEBUG 0

static char oma = 'x', opponent = 'o';
static class map 
{
    vector<char> contents;
    vector<unsigned> freeheight;
    typedef pair<unsigned, unsigned> paer;
    typedef list<paer> drophist_t;
    drophist_t drophist;
public:
    unsigned cols, rows;
public:
    inline char At(unsigned x, unsigned y) const
    {
        return contents[x + y*cols];
    }
    
    bool Drop(unsigned col, char mark);
    void Undrop();
    
    void Insert(const char *s)
    {
        while(*s && *s!='|')++s;
        unsigned colcount = 0;
        for(const char *p = s; *p == '|' && p[4] == '|'; p+=4)++colcount;
        
        cols = colcount;
        
        contents.reserve(contents.size() + colcount);
        
        for(unsigned a=0; a<colcount; ++a)
        {
            char c = s[2];
            if(c=='X') { c = opponent = 'x'; oma = 'o'; }
            else if(c=='O') { c = opponent = 'o'; oma = 'x'; }
            s += 4;
            contents.push_back(c);
        }
        ++rows;
        
        freeheight.resize(cols);
        for(unsigned col=0; col<cols; ++col)
        {
            unsigned free=0;
            for(unsigned r=0; r<rows; ++r)
            {
                if(contents[col + r*cols] != ' ')break;
                ++free;
            }
            freeheight[col] = free;
        }
    }
    
    char CheckWinner() const;
    void PrintField() const;
    
} map;

char map::CheckWinner() const
{
    for(unsigned x=0; x<cols; ++x)
        for(unsigned y=0; y<rows; ++y)
        {
            if(x < cols-3)
            {
                /* Check the horizontal */
                const char *const cptr = &contents[x+y*cols];
                if(*cptr == ' ')continue;
                if(*cptr == cptr[1] && *cptr == cptr[2] && *cptr == cptr[3])return *cptr;
                if(y < rows-3)
                {
                    /* And one diagonal */
                    if(*cptr == cptr[1+cols]
                    && *cptr == cptr[2+cols*2]
                    && *cptr == cptr[3+cols*3])return *cptr;
                }
                /* And another diagonal */
                char c = cptr[3];
                if(c == cptr[2+cols] && c == cptr[1+cols*2] && c == cptr[cols*3])return c;
            }
            if(y < rows-3)
            {
                /* And the vertical. */
                const char *const cptr = &contents[x+y*cols];
                if(*cptr == ' ')continue;
                if(*cptr == cptr[cols] && *cptr == cptr[cols*2] && *cptr == cptr[cols*3])return *cptr;
            }
        }
    return 0;
}

void map::PrintField() const
{
    printf("\33[32m  ");
    for(unsigned a=0; a<cols; ++a)printf("%c---", a?'-':',');
    printf(".\n");
    for(unsigned a=0; a<rows; ++a)
    {
        printf("  ");
        for(unsigned b=0; b<cols; ++b)
        {
            char c = At(b, a);
            printf("| \33[1;%dm%c \33[0;32m", c==oma ? 36 : 30, c);
        }
        printf("|");
        if(a == 1)printf("\33[34;1m  You: %c\33[32m", oma);
        else if(a == 2)printf("\33[34;1m  Foe: %c\33[32m", opponent);
        printf("\n");
    }
    printf("  ");
    for(unsigned a=0; a<cols; ++a)printf("%c---", a?'-':'`');
    printf("'\33[31m\n ");
    for(unsigned a=0; a<cols; ++a)printf("%4d", a);
    printf("\33[m\n");
    char c = CheckWinner();
    if(c)printf("Winner: '%c'\n", c);
}

bool map::Drop(unsigned col, char mark)
{
    unsigned f = freeheight[col];
    if(!f)return false;
    
    unsigned r = --freeheight[col];
    
    contents[col + r*cols] = mark;
    drophist.push_front(paer(col, r));
    
    return true;
}

void map::Undrop()
{
    if(drophist.size())
    {
        drophist_t::iterator i = drophist.begin();
        unsigned col = i->first;
        unsigned r   = i->second;
        contents[col + r*cols] = ' ';
        drophist.erase(i);
        
        ++freeheight[col];
    }
}

static int reclimit = 2;

class lohi
{
    int good, bad;
    int wincount, losecount;
public:
    lohi() : good(0), bad(0), wincount(0), losecount(0) { }
    lohi(int g, int b) : good(g), bad(b), wincount(0), losecount(0) { }
    void improve(const lohi &b)
    {
        if(b.good > good)good = b.good;
        if(b.bad  > bad)bad = b.bad;
        wincount += b.wincount;
        losecount+= b.losecount;
    }
    void setgood(int g) { good = g; } inline int getgood() const { return good; }
    void setbad(int b) { bad = b; }   inline int getbad() const { return bad; }
    void addwin() { ++wincount; }
    void addlose() { ++losecount; }
    int canwin() const { return wincount; }
    int canlose() const { return losecount; }
    
    operator bool () const { return good || bad; }
};

#if HISTORYDEBUG
static list<unsigned> selstack;
#endif

static lohi MoveScore(char good,char bad, unsigned col, int round=0)
{
    char thisturn = (round&1) ? opponent : oma;
    char nextturn = (round&1) ? oma : opponent;
    if(!map.Drop(col, thisturn))
    {
        return lohi(-1, -1);
    }
    
#if HISTORYDEBUG
    selstack.push_front(col);
#endif
    
    lohi result(1, 0);
    
    char c = map.CheckWinner();
    if(c == bad)
    {
        /* Lose! Losing is bad. */
        result.setbad(3 * (1 + reclimit - round));
        result.addlose();
        printf("%*sLosing.\n", round+1,"");
    }
    else if(c == good)
    {
        /* Good! The farther, the less sure though. */
        result.setgood(3 * (1 + reclimit - round));
        result.addwin();
        printf("%*sWinning.\n", round+1,"");
    }
    else
    {
        map.Undrop();
        map.Drop(col, nextturn);
        char c2 = map.CheckWinner();
        map.Undrop();
        map.Drop(col, thisturn);
#if 0
        if(c2 == good)             /* Jos voisi saada hyvinkin */
        {
            //printf("Col %u at round %u should not be given for bad.\n", col, round);
            result.setbad(1 + (reclimit - round));
            printf("%*sWanting.\n", round+1,"");
        }
        else
#endif
        if(c != bad && c2 == bad)
        {
            /* This is essentially important to _not_ give */
            //map.PrintField();
            //printf("Col %u at round %u is essential to not give!\n", col, round);
            result.setgood(1 + 2*(reclimit - round));
            printf("%*sKeeping.\n", round+1,"");
        }
        
        if(round < reclimit)
        {
            lohi mine;
            for(unsigned a=0; a<map.cols; ++a)
            {
                lohi test = MoveScore(good,bad, cols[a].col, round+1);
                if(thisturn == good)
                {
                    /* Enemy may do anything. */
                    result.improve(test);
                }
                else
                {
                    /* Enemy may do anything. */
                    result.improve(test);
                }
            }
        }
    }
    
    map.Undrop();
    
#if HISTORYDEBUG
    if(result)
    {
        printf("%*s", round+1,"");
        printf("Returning good=%d,bad=%d from round %d (cols", result.getgood(), result.getbad(), round);
        list<unsigned>::reverse_iterator i;
        for(i=selstack.rbegin(); i!=selstack.rend(); ++i)
            printf(" %u", *i);
        printf(")\n");
    }
    
    selstack.erase(selstack.begin());
#endif
    return result;
}

struct scorec
{
    int winchance, losechance;
    unsigned col;
};
static int Guess(char good, char bad, vector<scorec> &bests)
{
    int score;
    for(reclimit=3; reclimit<=9-limitdec; ++reclimit)
    {
        score = -32000;
        bests.clear();
        
        printf("Trying with recursion limit %d\n", reclimit);
        for(unsigned a=0; a<map.cols; ++a)
        {
            //if(a != 6)continue;
            
            lohi s = MoveScore(good,bad, a);
            printf(" Pos %u: Good=%d, bad=%d, winchans=%u, losechans=%u\n",
                   a, s.getgood(), s.getbad(),
                   s.canwin(), s.canlose());
            
            /* This column is full */
            if(s.getgood() == -1
            && s.getbad() == -1)continue;
            
            int sc = s.getgood() - s.getbad();
            if(!s.canwin() && s.canlose())continue;
            
            if(sc > score)
            {
                bests.clear();
                score = sc;
            }
            if(sc >= score)
            {
                scorec tmp;
                tmp.winchance = s.canwin();
                tmp.losechance= s.canlose();
                tmp.col = a;
                bests.push_back(tmp);
            }
        }
        if((score && bests.size() < map.cols * 6/7) || reclimit>=6)break;
    }
    return score;
}
int main(void)
{
    printf("It's your turn. Copypaste the field, thanks\n");
/*
Example field:
   ---------------------------
  |   |   |   | x |   |   |   |
  |   | o |   | x | o |   |   |
  |   | x |   | x | x |   |   |
  |   | x |   | o | o |   |   |
  | o | o |   | x | x |   |   |
  | o | o |   | x | o | x | O |
   --------------------------- 
    0   1   2   3   4   5   6
*/
    for(;;)
    {
        char Buf[512];
        if(!fgets(Buf, sizeof Buf, stdin))break;
        if(Buf[2]=='|' && Buf[6]=='|')map.Insert(Buf);
        if(strstr(Buf, "oved:"))break;
    }
    printf("Thanks.\n");
    map.PrintField();
    
    if(map.CheckWinner() == opponent)
    {
        printf("You lost!\n");
        return 0;
    }
    
    vector<scorec> good, bad;
    
    printf("Analyzing which moves would be good\n");
    int goodscore = Guess(oma, opponent, good);
    printf("Analyzing which moves would be bad\n");
    int badscore  = Guess(opponent, oma, bad);
    
    printf("\n");
    if(!good.size())
    {
        printf("You have lost. Try it if you don't believe.\n");
        return 0;
    }
    if(good.size() == 1)
        printf("Your logical choice is");
    else
        printf("Your move, pick one:");
    for(unsigned a=0; a<good.size(); ++a)
        printf(" %u", good[a].col);
    printf(". (score: %d)\n", goodscore);
    
    if(bad.size() == 1)
        printf("Whatever you do, DON'T move to");
    else
        printf("These moves are specially bad:");
    for(unsigned a=0; a<bad.size(); ++a)
        printf(" %u", bad[a].col);
    printf(". (score: %d)\n", -badscore);

    for(unsigned a=0; a<good.size(); ++a)
        for(unsigned b=0; b<bad.size(); ++b)
            if(good[a].col == bad[b].col)
            {
                printf("This must be disappointing...\n");
                goto NotHappySkip;
            }
NotHappySkip:
    if(badscore <= 0)printf("Your play looks promising.");
    else if(badscore > goodscore)printf("My opinion is that you will lose.\n");

    int wins=0, loss=0;
    for(unsigned a=0; a<good.size(); ++a)
    {
        wins += good[a].winchance;
        loss += good[a].losechance;
    }
    if(!wins && loss)printf("You will definitely lose.\n");
        
    return 0;
}
